@@ -61,7 +61,7 @@ struct ploop_check_desc {
 	__u32 *alloc_head;
 };
 
-static int read_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg)
+int read_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg)
 {
 	ssize_t res;
 
@@ -78,7 +78,7 @@ static int read_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg
 	return SYSEXIT_READ;
 }
 
-static int write_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg)
+int write_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg)
 {
 	ssize_t res;
 
@@ -98,7 +98,7 @@ static int fsync_safe(int fd)
 {
 	if (fsync(fd)) {
 		ploop_err(errno, "fsync");
-		return SYSEXIT_WRITE;
+		return SYSEXIT_FSYNC;
 	}
 	return 0;
 }
@@ -218,6 +218,12 @@ int ploop_check(char *img, int flags, int ro, int raw, int verbose, __u32 *block
 	int check = (flags & CHECK_DETAILED);
 	int version;
 
+	/* If an image is on read-only file system */
+	if (!ro && access(img, W_OK) == -1 && errno == EROFS) {
+		ploop_log(2, "Read-only file system for %s", img);
+		ro = 1; /* don't error out, do read-only check */
+	}
+
 	fd = open(img, ro ? O_RDONLY : O_RDWR);
 	if (fd < 0) {
 		ploop_err(errno, "ploop_check: can't open %s",
@@ -457,6 +463,7 @@ static int fill_hole(const char *image, int fd, off_t start, off_t end, int *log
 		if (!repair)
 			return SYSEXIT_PLOOPFMT;
 		*log = 1;
+		print_output(0, "filefrag -vs", image);
 		ploop_log(0, "Reallocating sparse blocks back");
 	}
 
@@ -533,6 +540,12 @@ static int check_and_repair_sparse(const char *image, int fd, __u64 cluster, int
 		for (i = 0; i < fiemap->fm_mapped_extents; i++) {
 			if (fm_ext[i].fe_flags & FIEMAP_EXTENT_LAST)
 				last = 1;
+
+			if (fm_ext[i].fe_logical >= st.st_size) {
+				last = 1;
+				break;
+			}
+
 			if ((fm_ext[i].fe_flags & FIEMAP_EXTENT_UNWRITTEN) &&
 			    (fm_ext[i].fe_logical % cluster ||
 					fm_ext[i].fe_length % cluster)) {
@@ -562,26 +575,26 @@ static int check_and_repair_sparse(const char *image, int fd, __u64 cluster, int
 			fill_hole(image, fd, prev_end, st.st_size, &log, repair))
 		return SYSEXIT_PLOOPFMT;
 
+	if (log)
+		print_output(0, "filefrag -vs", image);
+
 	return 0;
 }
 
 int check_deltas(struct ploop_disk_images_data *di, char **images,
-		struct ploop_mount_param *param, int raw, __u32 *blocksize)
+		int raw, __u32 *blocksize)
 {
 	int i;
 	int ret = 0;
 
 	for (i = 0; images[i] != NULL; i++) {
-		int ro;
 		int flags = CHECK_DETAILED | (di ? (CHECK_DROPINUSE | CHECK_REPAIR_SPARSE) : 0);
 		int raw_delta = (raw && i == 0);
 		__u32 cur_blocksize = raw_delta ? *blocksize : 0;
 
-		ro = (images[i+1] != NULL || param->ro) ? 1 : 0;
-		ret = ploop_check(images[i], flags, ro, raw_delta, 0, &cur_blocksize);
+		ret = ploop_check(images[i], flags, 0, raw_delta, 0, &cur_blocksize);
 		if (ret) {
-			ploop_err(0, "%s (%s): irrecoverable errors",
-					images[i], ro ? "ro" : "rw");
+			ploop_err(0, "%s : irrecoverable errors", images[i]);
 			break;
 		}
 		if (*blocksize == 0)
@@ -596,3 +609,55 @@ int check_deltas(struct ploop_disk_images_data *di, char **images,
 
 	return ret;
 }
+
+int check_deltas_same(const char *img1, const char *img2) {
+	int i, ret;
+	int fd[2] = { -1, -1 };
+	struct ploop_pvd_header vh[2];
+	struct stat st[2];
+	const char *img[2] = { img1, img2 };
+
+	for (i = 0; i < 2; i++) {
+		fd[i] = open(img[i], O_RDONLY);
+		if (fd[i] < 0) {
+			ploop_err(errno, "Can't open %s", img[i]);
+			ret = SYSEXIT_OPEN;
+			goto done;
+		}
+
+		if (fstat(fd[i], &st[i]) < 0) {
+			ploop_err(errno, "Can't fstat %s", img[i]);
+			ret = SYSEXIT_FSTAT;
+			goto done;
+		}
+
+		ret = read_safe(fd[i], &vh[i], sizeof(vh[i]), 0,
+				"read PVD header");
+		if (ret)
+			goto done;
+		/* clear "in use" flag since it might be different */
+		vh[i].m_DiskInUse = 0;
+	}
+
+	ret = SYSEXIT_PLOOPFMT;
+
+	if (st[0].st_size != st[1].st_size) {
+		ploop_err(0, "Images %s and %s are of different size",
+				img[0], img[1]);
+		goto done;
+	}
+
+	if (memcmp(&vh[0], &vh[1], sizeof(vh[0])) != 0) {
+		ploop_err(0, "Images %s and %s are not identical",
+				img[0], img[1]);
+		goto done;
+	}
+
+	ret = 0;
+done:
+	for (i = 0; i < 2; i++)
+		if (fd[i] >= 0)
+			close(fd[i]);
+
+	return ret;
+}
