@@ -21,7 +21,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <limits.h>
@@ -29,14 +28,51 @@
 
 #include "ploop.h"
 
+#define __stringify_1(x...)	#x
+#define __stringify(x...)	__stringify_1(x)
+
+/* A macro to create a list of versions of an e2fs utility
+ * (such as tune2fs, resize2fs or dumpe2fs) to look for,
+ * ordered by priority:
+ *  1 Our own private version from /usr/libexec
+ *  2 A version from e4fsprogs (for RHEL5 systems)
+ *  3 A standard one from /sbin
+ *  4 Just a name without a path, letting run_prg()
+ *    to look it up in all the standard locations.
+ */
+#define GEN_E2FS_PROG(name) \
+static char * name ## 2fs_progs[] = {			\
+	"/usr/libexec/"	__stringify(name) "2fs",	\
+	"/sbin/"	__stringify(name) "4fs",	\
+	"/sbin/"	__stringify(name) "2fs",	\
+	__stringify(name) "2fs",			\
+	NULL};						\
+
+GEN_E2FS_PROG(tune)
+GEN_E2FS_PROG(resize)
+GEN_E2FS_PROG(dumpe)
+
+#undef GEN_E2FS_PROG
+
+static char *get_prog(char *progs[])
+{
+	int i;
+
+	for (i = 0; progs[i] != NULL; i++)
+		if (access(progs[i], X_OK) == 0)
+			return progs[i];
+	/* the last in the list is default */
+	return progs[i - 1];
+}
+
 int create_gpt_partition(const char *device, off_t size, __u32 blocksize)
 {
 	unsigned long long start = blocksize;
-	unsigned long long end = size - start - GPT_DATA_SIZE;
+	unsigned long long end = (size - blocksize) / blocksize * blocksize;
 	char *argv[7];
 	char s1[22], s2[22];
 
-	if (size <= start + GPT_DATA_SIZE) {
+	if (size <= start + blocksize) {
 		ploop_err(0, "Image size should be greater than %llu", start);
 		return -1;
 	}
@@ -44,9 +80,9 @@ int create_gpt_partition(const char *device, off_t size, __u32 blocksize)
 	argv[1] = "-s";
 	argv[2] = (char *)device;
 	argv[3] = "mklabel gpt mkpart primary";
-	snprintf(s1, sizeof(s1), "%llus", (start >> 3) << 3);
+	snprintf(s1, sizeof(s1), "%llub", start << PLOOP1_SECTOR_LOG);
 	argv[4] = s1;
-	snprintf(s2, sizeof(s2), "%llus", ((end >> 3) << 3) - 1);
+	snprintf(s2, sizeof(s2), "%llub", (end << PLOOP1_SECTOR_LOG)-1);
 	argv[5] = s2;
 	argv[6] = NULL;
 
@@ -62,8 +98,9 @@ int make_fs(const char *device, const char *fstype, unsigned int fsblocksize)
 {
 	char part_device[64];
 	char fsblock_size[14];
-	char *argv[8];
+	char *argv[10];
 	char ext_opts[1024];
+	__u64 max_online_resize;
 
 	fsblocksize = fsblocksize != 0 ? fsblocksize : 4096;
 
@@ -77,19 +114,32 @@ int make_fs(const char *device, const char *fstype, unsigned int fsblocksize)
 	snprintf(fsblock_size, sizeof(fsblock_size), "-b%u",
 			fsblocksize);
 	argv[4] = fsblock_size;
-	/* Reserve enough space so that the block group descriptor table can grow to 16T */
-	snprintf(ext_opts, sizeof(ext_opts), "-Elazy_itable_init,resize=4294967295");
+	/* Reserve enough space so that the block group descriptor table can grow to 16T
+	 * Note: the max_online_resize is __u32 in mkfs.ext4
+	 */
+	max_online_resize = PLOOP_MAX_FS_SIZE / fsblocksize;
+	if (max_online_resize > (__u32)~0)
+		max_online_resize = (__u32)~0;
+	snprintf(ext_opts, sizeof(ext_opts), "-Elazy_itable_init,resize=%llu",
+			 max_online_resize);
 	argv[5] = ext_opts;
-	argv[6] = part_device;
-	argv[7] = NULL;
+	/* Set the journal size to 128M to allow online resize up to 16T
+	 * independly on the initial image size
+	*/
+	argv[6] = "-Jsize=128";
+	argv[7] = "-i16384"; /* 1 inode per 16K disk space */
+	argv[8] = part_device;
+	argv[9] = NULL;
 
 	if (run_prg(argv))
 		return SYSEXIT_MKFS;
 
-	argv[0] = "tune2fs";
+	argv[0] = get_prog(tune2fs_progs);
 	argv[1] =  "-ouser_xattr,acl";
-	argv[2] = part_device;
-	argv[3] = NULL;
+	argv[2] = "-c0";
+	argv[3] = "-i0";
+	argv[4] = part_device;
+	argv[5] = NULL;
 
 	if (run_prg(argv))
 		return SYSEXIT_MKFS;
@@ -108,13 +158,13 @@ void tune_fs(const char *target, const char *device, unsigned long long size_sec
 		ploop_err(errno, "tune_fs: can't statfs %s", target);
 		return;
 	}
-	reserved_blocks = size_sec / 100 * 5 * 512 / fs.f_bsize;
+	reserved_blocks = size_sec / 100 * 5 * SECTOR_SIZE / fs.f_bsize;
 	if (reserved_blocks == 0) {
 		ploop_err(0, "Can't set reserved blocks for size %llu",
 				size_sec);
 		return;
 	}
-	argv[0] = "tune2fs";
+	argv[0] = get_prog(tune2fs_progs);
 	argv[1] = "-r";
 	snprintf(buf, sizeof(buf), "%llu", reserved_blocks);
 	argv[2] = buf;
@@ -124,27 +174,12 @@ void tune_fs(const char *target, const char *device, unsigned long long size_sec
 	run_prg(argv);
 }
 
-static char *get_resize_prog(void)
-{
-	int i;
-	struct stat st;
-	static char *progs[] = {"/usr/libexec/resize2fs", "/sbin/resize4fs", "/sbin/resize2fs", NULL};
-
-	for (i = 0; progs[i] != NULL; i++)
-		if (stat(progs[i], &st) == 0)
-			return progs[i];
-
-	return "resize2fs";
-}
-
 int resize_fs(const char *device, off_t size_sec)
 {
-	char *prog;
 	char *argv[5];
 	char buf[22];
 
-	prog = get_resize_prog();
-	argv[0] = prog;
+	argv[0] = get_prog(resize2fs_progs);
 	argv[1] = "-p";
 	argv[2] = (char *)device;
 	if (size_sec) {
@@ -177,7 +212,8 @@ int dumpe2fs(const char *device, struct dump2fs_data *data)
 	FILE *fp;
 	int found = BLOCK_COUNT_BIT | BLOCK_FREE_BIT | BLOCK_SIZE_BIT;
 
-	snprintf(cmd, sizeof(cmd),  "LANG=C /sbin/dumpe2fs -h %s", device);
+	snprintf(cmd, sizeof(cmd), "LANG=C " DEF_PATH_ENV " %s -h %s",
+			get_prog(dumpe2fs_progs), device);
 	fp = popen(cmd, "r");
 	if (fp == NULL) {
 		ploop_err(0, "Failed %s", cmd);
@@ -208,13 +244,13 @@ int dumpe2fs(const char *device, struct dump2fs_data *data)
 	return 0;
 }
 
-int e2fsck(const char *device, int flags)
+int e2fsck(const char *device, int flags, int *rc)
 {
 	char *arg[5];
 	int i = 0;
 	int ret;
 
-	arg[i++] = "e2fsck";
+	arg[i++] = "fsck.ext4";
 	if (flags & E2FSCK_PREEN)
 		arg[i++] = "-p";
 	if (flags & E2FSCK_FORCE)
@@ -225,6 +261,9 @@ int e2fsck(const char *device, int flags)
 	if (run_prg_rc(arg, &ret))
 		return SYSEXIT_FSCK;
 
+	if (rc)
+		*rc = ret;
+
 	/* exit code < 4 is OK, see man e2fsck */
 	if (ret >= 4) {
 		ploop_err(0, "e2fsck failed (exit code %d)\n", ret);
