@@ -119,37 +119,42 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 		if (find_delta_names(device, end_level, end_level,
 				     &src_image, &fmt)) {
 			ploop_err(errno, "find_delta_names");
-			return(SYSEXIT_SYSFS);
+			ret = SYSEXIT_SYSFS;
+			goto done;
 		}
 
 		if ((ret = read_size_from_image(src_image, 0, &src_size)))
-			return ret;
+			goto done;
 	}
 
 	if (find_delta_names(device, start_level, start_level,
 			     &dst_image, &fmt)) {
 		ploop_err(errno, "find_delta_names");
-		return(SYSEXIT_SYSFS);
+		ret = SYSEXIT_SYSFS;
+		goto done;
 	}
 
 	if (strcmp(fmt, "raw") == 0)
 		dst_is_raw = 1;
 
 	if ((ret = read_size_from_image(dst_image, dst_is_raw, &dst_size)))
-		return ret;
+		goto done;
 
-	if (src_size <= dst_size)
-		return 0;
+	if (src_size <= dst_size) {
+		ret = 0;
+		goto done;
+	}
 
 	if (dst_is_raw) {
-		return grow_raw_delta(dst_image, S2B(src_size - dst_size));
+		ret = grow_raw_delta(dst_image, S2B(src_size - dst_size));
+		goto done;
 	}
 
 	/* Here we know for sure that destination delta is in ploop1 format */
-
 	if (open_delta(&odelta, dst_image, O_RDWR, OD_NOFLAGS)) {
 		ploop_err(errno, "open_delta");
-		return(SYSEXIT_OPEN);
+		ret = SYSEXIT_OPEN;
+		goto done;
 	}
 
 	if (dirty_delta(&odelta)) {
@@ -220,6 +225,8 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 	}
 
 done:
+	free(src_image);
+	free(dst_image);
 	close_delta(&odelta);
 	return ret;
 }
@@ -606,6 +613,63 @@ merge_done2:
 	return ret;
 }
 
+static const char *get_devs_str(char **devs, char *buf, int size)
+{
+	char **p;
+	char *sp = buf;
+	char *ep = buf + size;
+
+	for (p = devs; *p != NULL; p++) {
+		sp += snprintf(sp, ep - sp, "%s ", *p);
+		if (sp >= ep)
+			break;
+	}
+	return buf;
+}
+
+static int check_snapshot_mount(struct ploop_disk_images_data *di,
+		int temporary, const char *parent_fname,
+		const char *child_fname, const char *child_guid)
+{
+	int ret = 0;
+	char **devs, **p;
+	char buf[512];
+
+	/* Check if upper (child) delta that will be merged and
+	 * destroyed is mounted
+	 */
+	if (guidcmp(child_guid, di->top_guid) != 0 &&
+			ploop_get_dev_by_delta(di->images[0]->file,
+				child_fname, NULL, &devs) == 0)
+	{
+		ploop_err(0, "Snapshot is busy by device(s): %s",
+				get_devs_str(devs, buf, sizeof(buf)));
+		ret = SYSEXIT_EBUSY;
+		goto out;
+	}
+
+	/* check if snapshot (parent) delta is mounted */
+	if (ploop_get_dev_by_delta(di->images[0]->file,
+			parent_fname, NULL, &devs) == 0)
+	{
+		/* unmount temporary snapshot */
+		if (temporary) {
+			for (p = devs; *p != NULL; p++) {
+				ret = ploop_umount(*p, NULL);
+				if (ret)
+					break;
+			}
+		} else {
+			ploop_err(0, "Snapshot is busy by device(s): %s",
+					get_devs_str(devs, buf, sizeof(buf)));
+			ret = SYSEXIT_EBUSY;
+		}
+	}
+out:
+	ploop_free_array(devs);
+	return ret;
+}
+
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode)
 {
 	char conf[PATH_MAX];
@@ -625,6 +689,7 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 	int raw = (di->mode == PLOOP_RAW_MODE);
 	int online = 0;
 	int snap_idx;
+	int temporary = 0;
 	struct merge_info info = {};
 	int i, nelem;
 
@@ -635,6 +700,9 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 				guid);
 		return ret;
 	}
+
+	temporary = di->snapshots[snap_idx]->temporary;
+
 	fname = find_image_by_guid(di, guid);
 	if (fname == NULL) {
 		ploop_err(0, "Can't find image by uuid %s",
@@ -656,6 +724,7 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 					child_guid);
 			goto err;
 		}
+		temporary = di->snapshots[snap_idx]->temporary;
 	} else if (merge_mode == PLOOP_MERGE_WITH_PARENT) {
 		parent_guid = di->snapshots[snap_idx]->parent_guid;
 		child_guid = (char *)guid;
@@ -670,6 +739,14 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 					 parent_guid);
 			goto err;
 		}
+		snap_idx = find_snapshot_by_guid(di, parent_guid);
+		if (snap_idx == -1) {
+			ploop_err(0, "Can't find snapshot by uuid %s",
+					parent_guid);
+
+			goto err;
+		}
+		temporary = di->snapshots[snap_idx]->temporary;
 	} else
 		assert(0);
 
@@ -680,7 +757,7 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 		return SYSEXIT_PARAM;
 	}
 
-	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
+	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
 	if (ret == -1)
 		goto err;
 	else if (ret == 0)
@@ -744,6 +821,11 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 	}
 	names[2] = NULL;
 
+	ret = check_snapshot_mount(di, temporary, parent_fname,
+			child_fname, child_guid);
+	if (ret)
+		goto err;
+
 	/* make validation before real merge */
 	ret = ploop_di_merge_image(di, child_guid, &delete_fname);
 	if (ret)
@@ -763,8 +845,8 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 		ret = SYSEXIT_UNLINK;
 	}
 	if (ret == 0)
-		ploop_log(0, "ploop snapshot %s has been successfully merged",
-				guid);
+		ploop_log(0, "ploop %s %s has been successfully merged",
+				get_snap_str(temporary), guid);
 
 err:
 	for (i = 0; names[i] != NULL; i++)
@@ -781,7 +863,7 @@ int ploop_merge_snapshot(struct ploop_disk_images_data *di, struct ploop_merge_p
 	int ret = SYSEXIT_PARAM;
 	const char *guid = NULL;
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
 	if (param->guid != NULL)
@@ -798,7 +880,7 @@ int ploop_merge_snapshot(struct ploop_disk_images_data *di, struct ploop_merge_p
 				break;
 		}
 	}
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
