@@ -96,7 +96,7 @@ int get_list_size(char **list)
 	return i;
 }
 
-static char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse)
+char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse)
 {
 	int n;
 	char **images;
@@ -162,20 +162,6 @@ err:
 	return NULL;
 }
 
-static int get_snapshot_count(struct ploop_disk_images_data *di)
-{
-	int n;
-	char **images;
-
-	images = make_images_list(di, di->top_guid, 1);
-	if (images == NULL)
-		return -1;
-	n = get_list_size(images);
-	free_images_list(images);
-
-	return n;
-}
-
 void free_images_list(char **images)
 {
 	int i;
@@ -261,7 +247,8 @@ static int default_fmt_version(void)
 	return ploop_is_large_disk_supported() ? PLOOP_FMT_V2 : PLOOP_FMT_V1;
 }
 
-static int check_size(unsigned long long sectors, __u32 blocksize, int version)
+static int do_check_size(unsigned long long sectors, __u32 blocksize, int version,
+		__u64 max_fs_size)
 {
 	__u64 max;
 
@@ -280,8 +267,8 @@ static int check_size(unsigned long long sectors, __u32 blocksize, int version)
 		return -1;
 	}
 
-	if (max > B2S(PLOOP_MAX_FS_SIZE))
-		max = B2S(PLOOP_MAX_FS_SIZE);
+	if (max_fs_size != 0 && max > B2S(max_fs_size))
+		max = B2S(max_fs_size);
 
 	if (sectors > max) {
 		ploop_err(0, "An incorrect block device size is specified: %llu sectors."
@@ -292,6 +279,11 @@ static int check_size(unsigned long long sectors, __u32 blocksize, int version)
 	return 0;
 }
 
+static int check_size(unsigned long long sectors, __u32 blocksize, int version)
+{
+	return do_check_size(sectors, blocksize, version, PLOOP_MAX_FS_SIZE);
+}
+
 int check_blockdev_size(unsigned long long sectors, __u32 blocksize, int version)
 {
 	if (check_size(sectors, blocksize, version))
@@ -307,8 +299,7 @@ int check_blockdev_size(unsigned long long sectors, __u32 blocksize, int version
 	return 0;
 }
 
-static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize,
-		int version)
+static int do_create_delta(const char *path, __u32 blocksize, off_t bdsize, int version)
 {
 	int fd;
 	void * buf = NULL;
@@ -324,12 +315,6 @@ static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize,
 		return -1;
 	}
 
-	if (version == PLOOP_FMT_UNDEFINED)
-		version = default_fmt_version();
-
-	if (check_blockdev_size(bdsize, blocksize, version))
-		return -1;
-
 	if (p_memalign(&buf, 4096, cluster))
 		return -1;
 
@@ -374,6 +359,27 @@ out_close:
 	return -1;
 }
 
+static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize, int version)
+{
+	if (check_blockdev_size(bdsize, blocksize, version))
+		return -1;
+
+	return do_create_delta(path, blocksize, bdsize, version);
+}
+
+int create_snapshot_delta(const char *path, __u32 blocksize, off_t bdsize,
+		int version)
+{
+	/* select version for new delta on top of RAW image */
+	if (version == PLOOP_FMT_UNDEFINED)
+		version = default_fmt_version();
+
+	if (do_check_size(bdsize, blocksize, version, 0))
+		return -1;
+
+	return do_create_delta(path, blocksize, bdsize, version);
+}
+
 static int create_empty_preallocated_delta(const char *path, __u32 blocksize,
 		off_t bdsize, int version)
 {
@@ -477,9 +483,9 @@ static int create_raw_delta(const char * path, off_t bdsize)
 		if (is_operation_cancelled())
 			goto out_close;
 		off_t copy = bdsize - pos;
-		if (copy > DEF_CLUSTER/512)
-			copy = DEF_CLUSTER/512;
-		if (WRITE(fd, buf, copy*512))
+		if (copy > DEF_CLUSTER/SECTOR_SIZE)
+			copy = DEF_CLUSTER/SECTOR_SIZE;
+		if (WRITE(fd, buf, copy*SECTOR_SIZE))
 			goto out_close;
 		pos += copy;
 	}
@@ -787,7 +793,7 @@ static int do_ioctl(int fd, int req)
 	return ret;
 }
 
-static int print_output(int level, const char *cmd, const char *arg)
+int print_output(int level, const char *cmd, const char *arg)
 {
 	FILE *fp;
 	char command[PATH_MAX];
@@ -1013,7 +1019,7 @@ int ploop_get_dev_by_mnt(const char *path, char *buf, int size)
 	return get_dev_by_mnt(path, 1, buf, size);
 }
 
-char *ploop_get_base_delta_uuid(struct ploop_disk_images_data *di)
+static char *get_base_delta_uuid(struct ploop_disk_images_data *di)
 {
 	int i;
 
@@ -1045,51 +1051,36 @@ int ploop_get_top_delta_fname(struct ploop_disk_images_data *di, char *out, int
 	return 0;
 }
 
-int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di,
-		int check_state, char *out, int len)
-{
-	int ret;
-	int running = 0;
-
-	if (di->nimages <= 0) {
-		ploop_err(0, "No images found in " DISKDESCRIPTOR_XML);
-		return -1;
-	}
-	ret = ploop_find_dev(di->runtime->component_name, di->images[0]->file,
-			out, len);
-	if (ret == 0 && check_state) {
-		if (ploop_get_attr(out, "running", &running)) {
-			ploop_err(0, "Can't get running attr for %s",
-					out);
-			return -1;
-		}
-		if (!running) {
-			ploop_err(0, "Unexpectedly found stopped ploop device %s",
-					out);
-			return -1;
-		}
-	}
-
-	return ret;
-}
-
 int ploop_get_dev(struct ploop_disk_images_data *di, char *out, int len)
 {
 	int ret;
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev(di->runtime->component_name, di->images[0]->file, out, len);
+	ret = ploop_find_dev_by_cn(di, di->runtime->component_name, 1, out, len);
 
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
 
+int ploop_is_mounted(struct ploop_disk_images_data *di)
+{
+	int ret;
+	char device[64];
+
+	if (di->nimages == 0 && ploop_read_dd(di))
+		return -1;
+
+	ret = ploop_find_dev_by_dd(di, device, sizeof(device));
+
+	return (ret == -1 ? -1 : !ret);
+}
+
 int ploop_get_devs(struct ploop_disk_images_data *di, char ***out)
 {
-	return ploop_get_dev_by_delta(di->images[0]->file, NULL, out);
+	return ploop_get_dev_by_delta(di->images[0]->file, NULL, NULL, out);
 }
 
 static int reread_part(const char *device)
@@ -1127,7 +1118,7 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 		return SYSEXIT_MOUNT;
 
 	if (param->fsck && (strncmp(fstype, "ext", 3) == 0))
-		if (e2fsck(part_device, E2FSCK_PREEN))
+		if (e2fsck(part_device, E2FSCK_PREEN, &param->fsck_rc))
 			return SYSEXIT_FSCK;
 
 	/* Two step mount
@@ -1163,6 +1154,15 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 	return 0;
 }
 
+static void print_sys_block_ploop(void)
+{
+	print_output(-1, "find",
+			"/sys/block/ploop[0-9]*/ -type f "
+			"-not -name '*event' "
+			"-not -name ve_device_add "
+			"| xargs tail | grep -v '^$'");
+}
+
 static int add_delta(int lfd, const char *image, struct ploop_ctl_delta *req)
 {
 	int fd;
@@ -1179,11 +1179,7 @@ static int add_delta(int lfd, const char *image, struct ploop_ctl_delta *req)
 
 	if (ioctl(lfd, PLOOP_IOC_ADD_DELTA, req) < 0) {
 		if (errno == EBUSY)
-			print_output(-1, "find",
-					"/sys/block/ploop[0-9]*/ -type f "
-					"-not -name '*event' "
-					"-exec echo {} \\; -exec cat {} \\;");
-
+			print_sys_block_ploop();
 		ploop_err(0, "Can't add image %s: %s", image,
 				(errno == ENOTSUP) ?
 					"unsupported underlying filesystem"
@@ -1198,6 +1194,184 @@ out:
 	return ret;
 }
 
+int replace_delta(const char *device, int level, const char *image)
+{
+	int fd = -1, lfd = -1;
+	int top_level = 0;
+	int ret;
+	__u32 blocksize = 0;
+	struct ploop_ctl_delta req = {};
+
+	fd = open(image, O_DIRECT | O_RDONLY);
+	if (fd < 0) {
+		ploop_err(errno, "Can't open file %s", image);
+		return SYSEXIT_OPEN;
+	}
+
+	lfd = open(device, O_RDONLY);
+	if (lfd < 0) {
+		ploop_err(errno, "Can't open device %s", device);
+		ret = SYSEXIT_DEVICE;
+		goto out;
+	}
+	if (ploop_get_attr(device, "top", &top_level)) {
+		ret = SYSEXIT_SYSFS;
+		goto out;
+	}
+
+	if (level < 0 || level >= top_level) {
+		ploop_err(0, "Invalid level %d specified, allowed values "
+				"are 0 to %d", level, top_level - 1);
+		ret = SYSEXIT_PARAM;
+		goto out;
+	}
+
+	if (ploop_get_attr(device, "block_size", (int*) &blocksize)) {
+		ret = SYSEXIT_SYSFS;
+		goto out;
+	}
+
+	req.c.pctl_cluster_log = ffs(blocksize) - 1;
+	req.c.pctl_level = level;
+	req.c.pctl_chunks = 1;
+	req.c.pctl_format = PLOOP_FMT_PLOOP1;
+	req.c.pctl_flags = PLOOP_FMT_RDONLY;
+
+	req.f.pctl_type = PLOOP_IO_AUTO;
+	req.f.pctl_fd = fd;
+
+	if (ioctl(lfd, PLOOP_IOC_REPLACE_DELTA, &req) < 0) {
+		if (errno == EBUSY)
+			print_sys_block_ploop();
+		ploop_err(errno, "Can't replace image %s", image);
+		ret = SYSEXIT_DEVIOC;
+		goto out;
+	}
+	ret = 0;
+
+out:
+	if (lfd >= 0)
+		close(lfd);
+	if (fd >= 0)
+		close(fd);
+
+	return ret;
+}
+
+int ploop_replace_image(struct ploop_disk_images_data *di,
+		struct ploop_replace_param *param)
+{
+	char dev[PATH_MAX];
+	char *file = NULL, *oldfile;
+	char conf[PATH_MAX], conf_tmp[PATH_MAX] = "";
+	int ret, level;
+
+	if (ploop_lock_dd(di))
+		return SYSEXIT_LOCK;
+
+	ret = SYSEXIT_PARAM;
+
+	if (!param || !param->file) {
+		ploop_err(0, "New image file not specified");
+		goto err;
+	}
+
+	file = realpath(param->file, NULL);
+	if (file == NULL) {
+		ploop_err(errno, "Error in realpath(%s)", param->file);
+		goto err;
+	}
+
+	if (ploop_find_dev_by_dd(di, dev, sizeof(dev))) {
+		ploop_err(0, "Can't find running ploop device");
+		goto err;
+	}
+
+	/* Image to be replaced is specified by either guid or level
+	 */
+	if (param->guid) {
+		if (!is_valid_guid(param->guid)) {
+			ploop_err(0, "Invalid guid specified: %s", param->guid);
+			goto err;
+		}
+
+		level = find_image_idx_by_guid(di, param->guid);
+		if (level == -1) {
+			ploop_err(0, "Can't find image by guid %s", param->guid);
+			goto err;
+		}
+	}
+	else if (param->cur_file) {
+		level = find_level_by_delta(dev, param->cur_file);
+		if (level < 0) {
+			ploop_err(0, "Can't find level by delta file name %s",
+					param->cur_file);
+			goto err;
+		}
+	}
+	else {
+		level = param->level;
+	}
+	/* Proper level check (against top_level) is to be done later
+	 * in replace_delta(). Here is just some basic sanity check.
+	 * */
+	if (level < 0 || level >= di->nimages) {
+		ploop_err(0, "Invalid level %d", level);
+		goto err;
+	}
+
+	/* check a new image */
+	ret = ploop_check(file, CHECK_DETAILED, 1, 0, 0, NULL);
+	if (ret)
+		return ret;
+
+	/* check that images are identical */
+	oldfile = param->cur_file ? : di->images[level]->file;
+	ret = check_deltas_same(file, oldfile);
+	if (ret)
+		return ret;
+
+	/* Write new dd.xml with changed image file */
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	oldfile = di->images[level]->file;
+	di->images[level]->file = file;
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
+	di->images[level]->file = oldfile;
+	if (ret)
+		goto err;
+
+	/* Do replace */
+	ploop_log(0, "Replacing %s with %s", di->images[level]->file, file);
+	ret = replace_delta(dev, level, file);
+	if (ret)
+		goto err;
+
+	/* Put a new dd.xml */
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s to %s", conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+		/* FIXME: how to rollback now? */
+		goto err;
+	}
+	conf_tmp[0] = '\0'; /* prevent unlink() below */
+
+	/* Change image in di */
+	free(di->images[level]->file);
+	di->images[level]->file = file; /* malloc()ed by realpath */
+	file = NULL; /* prevent free(file) below */
+
+	ret = 0;
+err:
+	if (file)
+		free(file);
+	if (conf_tmp[0])
+		unlink(conf_tmp);
+	ploop_unlock_dd(di);
+
+	return ret;
+}
+
 static int create_ploop_dev(int minor)
 {
 	char device[64];
@@ -1299,6 +1473,10 @@ static int add_deltas(struct ploop_disk_images_data *di,
 		goto err1;
 	}
 
+	ret = check_and_repair_gpt(param->device, blocksize);
+	if (ret)
+		goto err1;
+
 err1:
 	if (ret) {
 		int err = 0;
@@ -1405,7 +1583,7 @@ int ploop_mount(struct ploop_disk_images_data *di, char **images,
 	if (di && (ret = check_and_restore_fmt_version(di)))
 		goto err;
 
-	ret = check_deltas(di, images, param, raw, &blocksize);
+	ret = check_deltas(di, images, raw, &blocksize);
 	if (ret)
 		goto err;
 
@@ -1434,7 +1612,7 @@ err:
 	return ret;
 }
 
-static int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *param, int flags)
+int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *param, int flags)
 {
 	int ret;
 	char **images;
@@ -1489,12 +1667,12 @@ int ploop_mount_image(struct ploop_disk_images_data *di, struct ploop_mount_para
 	int ret;
 	char dev[64];
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
+	ret = ploop_find_dev_by_cn(di, di->runtime->component_name, 1, dev, sizeof(dev));
 	if (ret == -1) {
-		ploop_unlock_di(di);
+		ploop_unlock_dd(di);
 		return SYSEXIT_SYS;
 	}
 	if (ret == 0) {
@@ -1506,8 +1684,10 @@ int ploop_mount_image(struct ploop_disk_images_data *di, struct ploop_mount_para
 	}
 
 	ret = mount_image(di, param, 0);
+	if (ret == 0 && di->runtime->component_name == NULL)
+		merge_temporary_snapshots(di);
 err:
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
@@ -1570,39 +1750,34 @@ int ploop_umount_image(struct ploop_disk_images_data *di)
 	int ret;
 	char dev[PATH_MAX];
 
-	if (di->nimages == 0) {
-		ploop_err(0, "No images specified");
-		return SYSEXIT_PARAM;
-	}
-
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev_by_uuid(di, 0, dev, sizeof(dev));
+	ret = ploop_find_dev_by_cn(di, di->runtime->component_name, 0, dev, sizeof(dev));
 	if (ret == -1) {
-		ploop_unlock_di(di);
+		ploop_unlock_dd(di);
 		return SYSEXIT_SYS;
 	}
 	if (ret != 0) {
-		ploop_unlock_di(di);
+		ploop_unlock_dd(di);
 		ploop_err(0, "Image %s is not mounted", di->images[0]->file);
 		return SYSEXIT_DEV_NOT_MOUNTED;
 	}
 
 	ret = ploop_complete_running_operation(dev);
 	if (ret) {
-		ploop_unlock_di(di);
+		ploop_unlock_dd(di);
 		return ret;
 	}
 
 	ret = ploop_umount(dev, di);
 
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
 
-static int get_image_param_online(const char *device, off_t *size,
+int get_image_param_online(const char *device, off_t *size,
 		__u32 *blocksize, int *version)
 {
 	if (ploop_get_attr(device, "block_size",  (int *)blocksize))
@@ -1616,7 +1791,7 @@ static int get_image_param_online(const char *device, off_t *size,
 	return ploop_get_size(device, size);
 }
 
-static int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
+int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
 		off_t *size, __u32 *blocksize, int *version)
 {
 	int ret;
@@ -1666,7 +1841,7 @@ static int get_image_param_offline(struct ploop_disk_images_data *di, const char
 	return 0;
 }
 
-static int get_image_param(struct ploop_disk_images_data *di, const char *guid,
+int get_image_param(struct ploop_disk_images_data *di, const char *guid,
 		off_t *size, __u32 *blocksize, int *version)
 {
 	int ret;
@@ -1676,7 +1851,7 @@ static int get_image_param(struct ploop_disk_images_data *di, const char *guid,
 	 * get offline for non top delta.
 	 */
 	if (strcmp(di->top_guid, guid) == 0) {
-		ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
+		ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
 		if (ret == -1)
 			return SYSEXIT_SYS;
 		if (ret == 0)
@@ -1751,10 +1926,10 @@ int ploop_grow_image(struct ploop_disk_images_data *di, off_t size)
 	int ret;
 	char device[64];
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev_by_uuid(di, 1, device, sizeof(device));
+	ret = ploop_find_dev_by_dd(di, device, sizeof(device));
 	if (ret == -1) {
 		ret = SYSEXIT_SYS;
 		goto err;
@@ -1787,7 +1962,7 @@ int ploop_grow_image(struct ploop_disk_images_data *di, off_t size)
 	}
 
 err:
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
@@ -1799,13 +1974,12 @@ static int ploop_raw_discard(struct ploop_disk_images_data *di, const char *devi
 	char conf[PATH_MAX];
 	off_t new_end;
 
-	new_end = start + GPT_DATA_SIZE;
-	new_end = ROUNDUP(new_end, blocksize);
+	new_end = ROUNDUP(start, blocksize);
 
 	if (new_end >= end)
 		return 0;
 
-	ret = resize_gpt_partition(device, new_end);
+	ret = resize_gpt_partition(device, new_end, blocksize);
 	if (ret)
 		return ret;
 
@@ -1860,7 +2034,7 @@ static int shrink_device(struct ploop_disk_images_data *di,
 	ploop_log(0, "Offline shrink %s dev=%s size=%lu new_size=%lu, start=%u",
 			(raw) ? "raw" : "",
 			part_device, (long)part_dev_size, (long)new_size, part_start);
-	ret = e2fsck(part_device, E2FSCK_FORCE | E2FSCK_PREEN);
+	ret = e2fsck(part_device, E2FSCK_FORCE | E2FSCK_PREEN, NULL);
 	if (ret)
 		return ret;
 
@@ -1905,15 +2079,10 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 	int version;
 	off_t new_fs_size = 0;
 
-	if (di->nimages == 0) {
-		ploop_err(0, "No images in DiskDescriptor");
-		return SYSEXIT_DISKDESCR;
-	}
-
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev_by_uuid(di, 1, buf, sizeof(buf));
+	ret = ploop_find_dev_by_dd(di, buf, sizeof(buf));
 	if (ret == -1) {
 		ret = SYSEXIT_SYS;
 		goto err;
@@ -1978,23 +2147,27 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 	balloon_size = bytes2sec(st.st_size);
 
 	if (param->size == 0) {
-		int delta = 1024 * 1024;
+		__u64 delta = di->blocksize ?: 2048;
+		__u64 free_space;
 
-		/* Inflate balloon up to max free space */
+		/* Iteratively inflate balloon up to max free space */
 		if (statfs(mount_param.target, &fs) != 0) {
 			ploop_err(errno, "statfs(%s)", mount_param.target);
 			ret = SYSEXIT_FSTAT;
 			goto err;
 		}
-		if (fs.f_bfree <= delta / fs.f_bsize) {
-			ret = 0; // no free space
-			goto err;
-		}
 
-		new_balloon_size = balloon_size + B2S(fs.f_bfree * fs.f_bsize);
-		new_balloon_size -= B2S(delta);
-		ret = ploop_balloon_change_size(mount_param.device,
-				balloonfd, new_balloon_size);
+		free_space = B2S(fs.f_bfree * fs.f_bsize);
+
+		for (new_balloon_size = balloon_size + free_space;
+				delta < free_space && new_balloon_size > balloon_size;
+				delta *= 2)
+		{
+			ret = ploop_balloon_change_size(mount_param.device,
+					balloonfd, new_balloon_size - delta);
+			if (ret != SYSEXIT_FALLOCATE)
+				break;
+		}
 	} else if (new_size > dev_size) {
 		char conf[PATH_MAX];
 		char conf_tmp[PATH_MAX];
@@ -2013,7 +2186,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			ret = do_umount(mount_param.target);
 			if (ret)
 				goto err;
-			ret = e2fsck(part_device, E2FSCK_FORCE | E2FSCK_PREEN);
+			ret = e2fsck(part_device, E2FSCK_FORCE | E2FSCK_PREEN, NULL);
 			if (ret)
 				goto err;
 		}
@@ -2039,16 +2212,23 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			goto err;
 		}
 
-		ret = resize_gpt_partition(mount_param.device, 0);
+		ret = resize_gpt_partition(mount_param.device, 0, blocksize);
 		if (ret)
 			goto err;
 
 		/* resize up to the end of device */
-		ret = resize_fs(part_device, new_fs_size);
+		ret = resize_fs(part_device, 0);
 		if (ret)
 			goto err;
 	} else {
 		/* Grow or shrink fs but do not change block device size */
+		if (part_dev_size < new_fs_size) {
+			/* sync gpt with new_size */
+			ret = resize_gpt_partition(mount_param.device, new_size, blocksize);
+			if (ret)
+				goto err;
+		}
+
 		if (!mounted && param->offline_resize) {
 			/* Offline */
 			if (balloon_size != 0) {
@@ -2123,7 +2303,7 @@ err:
 		close(balloonfd);
 	if (mounted == 0)
 		ploop_umount(mount_param.device, di);
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 	free_mount_param(&mount_param);
 
 	return ret;
@@ -2287,16 +2467,15 @@ int ploop_convert_image(struct ploop_disk_images_data *di, int mode, int flags)
 	char conf[PATH_MAX];
 	int ret = -1;
 
+
+	if (ploop_lock_dd(di))
+		return SYSEXIT_LOCK;
+
 	if (di->mode == PLOOP_RAW_MODE) {
 		ploop_err(0, "Converting raw image is not supported");
-		return SYSEXIT_PARAM;
+		ret =  SYSEXIT_PARAM;
+		goto err;
 	}
-	if (di->nimages == 0) {
-		ploop_err(0, "No images specified");
-		return SYSEXIT_PARAM;
-	}
-	if (ploop_lock_di(di))
-		return SYSEXIT_LOCK;
 
 	di->mode = mode;
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
@@ -2324,7 +2503,7 @@ int ploop_convert_image(struct ploop_disk_images_data *di, int mode, int flags)
 	}
 
 err:
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
@@ -2448,7 +2627,7 @@ int ploop_change_fmt_version(struct ploop_disk_images_data *di, int new_version,
 		return SYSEXIT_PARAM;
 	}
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
 	if (di->mode == PLOOP_RAW_MODE) {
@@ -2457,7 +2636,7 @@ int ploop_change_fmt_version(struct ploop_disk_images_data *di, int new_version,
 		goto err;
 	}
 
-	rc = ploop_find_dev_by_uuid(di, 1, fname, sizeof(fname));
+	rc = ploop_find_dev_by_dd(di, fname, sizeof(fname));
 	if (rc == -1) {
 		ret = SYSEXIT_SYS;
 		goto err;
@@ -2503,7 +2682,7 @@ int ploop_change_fmt_version(struct ploop_disk_images_data *di, int new_version,
 	}
 
 	/* Recheck ploop state after locking */
-	rc = ploop_find_dev_by_uuid(di, 1, fname, sizeof(fname));
+	rc = ploop_find_dev_by_dd(di, fname, sizeof(fname));
 	if (rc == -1) {
 		ret = SYSEXIT_SYS;
 		goto err;
@@ -2547,7 +2726,7 @@ err_rm:
 
 err:
 	deinit_delta_array(&da);
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	if (ret == 0)
 		ploop_log(0, "ploop image has been successfully converted");
@@ -2664,7 +2843,7 @@ int check_and_restore_fmt_version(struct ploop_disk_images_data *di)
 	if (di->mode == PLOOP_RAW_MODE)
 		return 0;
 
-	if ((guid = ploop_get_base_delta_uuid(di)) == NULL ||
+	if ((guid = get_base_delta_uuid(di)) == NULL ||
 		 (base_id = find_image_idx_by_guid(di, guid)) == -1)
 	{
 		ploop_log(-1, "Unable to find base image");
@@ -2705,10 +2884,10 @@ static int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *
 	char dev[64];
 	int ret = -1;
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
+	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
 	if (ret == -1) {
 		ret = SYSEXIT_SYS;
 		goto err;
@@ -2732,7 +2911,7 @@ static int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *
 	}
 
 err:
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
@@ -2746,78 +2925,13 @@ int ploop_get_info_by_descr(const char *descr, struct ploop_info *info)
 	if (read_statfs_info(descr, info) == 0)
 		return 0;
 
-	ret = ploop_read_disk_descr(&di, descr);
+	ret = ploop_open_dd(&di, descr);
 	if (ret)
 		return ret;
 
 	ret = ploop_get_info(di, info);
 
-	ploop_free_diskdescriptor(di);
-
-	return ret;
-}
-
-static int do_snapshot(int lfd, int fd, struct ploop_ctl_delta *req)
-{
-	req->f.pctl_fd = fd;
-
-	if (ioctl(lfd, PLOOP_IOC_SNAPSHOT, req) < 0) {
-		ploop_err(errno, "PLOOP_IOC_SNAPSHOT");
-		return SYSEXIT_DEVIOC;
-	}
-
-	return 0;
-}
-
-int create_snapshot(const char *device, const char *delta, int syncfs)
-{
-	int ret;
-	int lfd = -1;
-	int fd = -1;
-	off_t bdsize;
-	struct ploop_ctl_delta req;
-	__u32 blocksize;
-	int version;
-
-	ret = ploop_complete_running_operation(device);
-	if (ret)
-		return ret;
-
-	ret = get_image_param_online(device, &bdsize,
-			&blocksize, &version);
-	if (ret)
-		return ret;
-
-	lfd = open(device, O_RDONLY);
-	if (lfd < 0) {
-		ploop_err(errno, "Can't open device %s", device);
-		return SYSEXIT_DEVICE;
-	}
-
-	fd = create_empty_delta(delta, blocksize, bdsize, version);
-	if (fd < 0) {
-		ret = SYSEXIT_OPEN;
-		goto err;
-	}
-
-	memset(&req, 0, sizeof(req));
-
-	req.c.pctl_format = PLOOP_FMT_PLOOP1;
-	req.c.pctl_flags = syncfs ? PLOOP_FLAG_FS_SYNC : 0;
-	req.c.pctl_cluster_log = ffs(blocksize) - 1;
-	req.c.pctl_size = 0;
-	req.c.pctl_chunks = 1;
-	req.f.pctl_type = PLOOP_IO_AUTO;
-
-	ploop_log(0, "Creating snapshot dev=%s img=%s", device, delta);
-	ret = do_snapshot(lfd, fd, &req);
-	if (ret)
-		unlink(delta);
-err:
-	if (lfd >= 0)
-		close(lfd);
-	if (fd >= 0)
-		close(fd);
+	ploop_close_dd(di);
 
 	return ret;
 }
@@ -2826,363 +2940,13 @@ int ploop_get_spec(struct ploop_disk_images_data *di, struct ploop_spec *spec)
 {
 	int ret;
 
-	if (ploop_lock_di(di))
+	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
 	ret = get_image_param(di, di->top_guid, &spec->size, &spec->blocksize,
 			&spec->fmt_version);
 
-	ploop_unlock_di(di);
-
-	return ret;
-}
-
-int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapshot_param *param)
-{
-	int ret;
-	int fd;
-	char dev[64];
-	char snap_guid[61];
-	char file_guid[61];
-	char fname[PATH_MAX];
-	char conf[PATH_MAX];
-	char conf_tmp[PATH_MAX];
-	int online = 0;
-	int n;
-	off_t size;
-	__u32 blocksize;
-	int version;
-
-	if (di->nimages == 0) {
-		ploop_err(0, "No images");
-		return SYSEXIT_PARAM;
-	}
-	if (param->guid != NULL && !is_valid_guid(param->guid)) {
-		ploop_err(0, "Incorrect guid %s", param->guid);
-		return SYSEXIT_PARAM;
-	}
-
-	if (ploop_lock_di(di))
-		return SYSEXIT_LOCK;
-
-	ret = gen_uuid_pair(snap_guid, sizeof(snap_guid),
-			file_guid, sizeof(file_guid));
-	if (ret) {
-		ploop_err(errno, "Can't generate uuid");
-		goto err_cleanup1;
-	}
-
-	if (param->guid != NULL) {
-		if (find_snapshot_by_guid(di, param->guid) != -1) {
-			ploop_err(0, "The snapshot %s already exist",
-				param->guid);
-			ret = SYSEXIT_PARAM;
-			goto err_cleanup1;
-		}
-		strcpy(snap_guid, param->guid);
-	}
-	n = get_snapshot_count(di);
-	if (n == -1) {
-		ret = SYSEXIT_PARAM;
-		goto err_cleanup1;
-	} else if (n > 128-2) {
-		/* The number of images limited by 128
-		   so the snapshot limit 128 - base_image - one_reserverd
-		 */
-		ret = SYSEXIT_PARAM;
-		ploop_err(errno, "Unable to create a snapshot."
-			" The maximum number of snapshots (%d) has been reached",
-			n-1);
-		goto err_cleanup1;
-	}
-
-	snprintf(fname, sizeof(fname), "%s.%s",
-			di->images[0]->file, file_guid);
-	ploop_di_change_guid(di, di->top_guid, snap_guid);
-	ret = ploop_di_add_image(di, fname, TOPDELTA_UUID, snap_guid);
-	if (ret)
-		goto err_cleanup1;
-
-	get_disk_descriptor_fname(di, conf, sizeof(conf));
-	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
-	ret = ploop_store_diskdescriptor(conf_tmp, di);
-	if (ret)
-		goto err_cleanup1;
-
-	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
-	if (ret == -1) {
-		ret = SYSEXIT_SYS;
-		goto err_cleanup2;
-	}
-	else if (ret == 0)
-		online = 1;
-
-	if (!online) {
-		// offline snapshot
-		ret = get_image_param_offline(di, snap_guid, &size, &blocksize, &version);
-		if (ret)
-			goto err_cleanup2;
-
-		fd = create_empty_delta(fname, blocksize, size, version);
-		if (fd < 0) {
-			ret = SYSEXIT_CREAT;
-			goto err_cleanup2;
-		}
-		close(fd);
-	} else {
-		// Always sync fs
-		ret = create_snapshot(dev, fname, 1);
-		if (ret)
-			goto err_cleanup2;
-	}
-
-	if (rename(conf_tmp, conf)) {
-		ploop_err(errno, "Can't rename %s %s",
-				conf_tmp, conf);
-		ret = SYSEXIT_RENAME;
-	}
-
-	if (ret && !online && unlink(fname))
-		ploop_err(errno, "Can't unlink %s",
-				fname);
-
-	ploop_log(0, "ploop snapshot %s has been successfully created",
-			snap_guid);
-err_cleanup2:
-	if (ret && !online && unlink(conf_tmp))
-		ploop_err(errno, "Can't unlink %s", conf_tmp);
-
-err_cleanup1:
-	ploop_unlock_di(di);
-
-	return ret;
-}
-
-int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
-		struct ploop_snapshot_switch_param *param)
-{
-	int ret;
-	int fd;
-	char dev[64];
-	char uuid[61];
-	char file_uuid[61];
-	char new_top_delta_fname[PATH_MAX] = "";
-	char *old_top_delta_fname = NULL;
-	char conf[PATH_MAX];
-	char conf_tmp[PATH_MAX];
-	off_t size;
-	const char *guid = param->guid;
-	int flags = param->flags;
-	__u32 blocksize;
-	int version;
-
-	if (!is_valid_guid(guid)) {
-		ploop_err(0, "Incorrect guid %s", guid);
-		return SYSEXIT_PARAM;
-	}
-
-	if (ploop_lock_di(di))
-		return SYSEXIT_LOCK;
-
-	ret = SYSEXIT_PARAM;
-	if (strcmp(di->top_guid, guid) == 0) {
-		ploop_err(errno, "Nothing to do, already on %s snapshot",
-				guid);
-		goto err_cleanup1;
-	}
-
-	if (find_snapshot_by_guid(di, guid) == -1) {
-		ploop_err(0, "Can't find snapshot by uuid %s",
-				guid);
-		goto err_cleanup1;
-	}
-
-	// Read image param from snapshot we going to switch on
-	ret = get_image_param(di, guid, &size, &blocksize, &version);
-	if (ret)
-		goto err_cleanup1;
-
-	ret = gen_uuid_pair(uuid, sizeof(uuid), file_uuid, sizeof(file_uuid));
-	if (ret) {
-		ploop_err(errno, "Can't generate uuid");
-		goto err_cleanup1;
-	}
-
-	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_DESTROY)) {
-		// device should be stopped
-		ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
-		if (ret == -1) {
-			ret = SYSEXIT_SYS;
-			goto err_cleanup1;
-		} else if (ret == 0) {
-			ret = SYSEXIT_PARAM;
-			ploop_err(0, "Unable to perform switch to snapshot operation"
-					" on running device (%s)",
-					dev);
-			goto err_cleanup1;
-		}
-		ret = ploop_di_remove_image(di, di->top_guid, 0, &old_top_delta_fname);
-		if (ret)
-			goto err_cleanup1;
-	} else if (param->guid_old != NULL) {
-		if (!is_valid_guid(param->guid_old)) {
-			ploop_err(0, "Incorrect guid %s", param->guid_old);
-			goto err_cleanup1;
-		}
-
-		if (find_snapshot_by_guid(di, param->guid_old) != -1) {
-			ploop_err(0, "Incorrect guid_old %s: already exists",
-					param->guid_old);
-			goto err_cleanup1;
-		}
-
-		ploop_di_change_guid(di, di->top_guid, param->guid_old);
-	}
-
-	if (flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE) {
-		ploop_di_change_guid(di, guid, TOPDELTA_UUID);
-	} else {
-		snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
-			di->images[0]->file, file_uuid);
-		ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
-		if (ret)
-			goto err_cleanup1;
-	}
-
-	get_disk_descriptor_fname(di, conf, sizeof(conf));
-	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
-	ret = ploop_store_diskdescriptor(conf_tmp, di);
-	if (ret)
-		goto err_cleanup1;
-
-	// offline snapshot
-	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE)) {
-		fd = create_empty_delta(new_top_delta_fname, blocksize, size, version);
-		if (fd == -1) {
-			ret = SYSEXIT_CREAT;
-			goto err_cleanup2;
-		}
-		close(fd);
-	}
-
-	if (rename(conf_tmp, conf)) {
-		ploop_err(errno, "Can't rename %s %s",
-				conf_tmp, conf);
-		ret = SYSEXIT_RENAME;
-		goto err_cleanup3;
-	}
-
-	/* destroy precached info */
-	drop_statfs_info(di->images[0]->file);
-
-	if (old_top_delta_fname != NULL) {
-		ploop_log(0, "Removing %s", old_top_delta_fname);
-		if (unlink(old_top_delta_fname))
-			ploop_err(errno, "Can't unlink %s",
-					old_top_delta_fname);
-	}
-
-	ploop_log(0, "ploop snapshot has been successfully switched");
-err_cleanup3:
-	if (ret && unlink(new_top_delta_fname))
-		ploop_err(errno, "Can't unlink %s",
-				conf_tmp);
-err_cleanup2:
-	if (ret && unlink(conf_tmp))
-		ploop_err(errno, "Can't unlink %s",
-				conf_tmp);
-err_cleanup1:
-	ploop_unlock_di(di);
-	free(old_top_delta_fname);
-
-	return ret;
-}
-
-int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, int flags)
-{
-	struct ploop_snapshot_switch_param param = {};
-
-	param.guid = (char *) guid;
-	param.flags = flags;
-
-	return ploop_switch_snapshot_ex(di, &param);
-}
-
-int ploop_delete_top_delta(struct ploop_disk_images_data *di)
-{
-	return ploop_delete_snapshot(di, di->top_guid);
-}
-
-/* delete snapshot by guid
- * 1) if guid is not active and last -> delete guid
- * 2) if guid is not last merge with child -> delete child
- */
-int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
-{
-	int ret;
-	char conf[PATH_MAX];
-	char *fname = NULL;
-	int nelem = 0;
-	char dev[64];
-	int snap_id;
-
-	if (ploop_lock_di(di))
-		return SYSEXIT_LOCK;
-
-	ret = SYSEXIT_PARAM;
-	snap_id = find_snapshot_by_guid(di, guid);
-	if (snap_id == -1) {
-		ploop_err(0, "Can't find snapshot by uuid %s",
-				guid);
-		goto err;
-	}
-	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
-	if (ret == -1) {
-		ret = SYSEXIT_SYS;
-		goto err;
-	}
-	else if (ret == 0 && strcmp(di->top_guid, guid) == 0) {
-		ret = SYSEXIT_PARAM;
-		ploop_err(0, "Unable to delete active snapshot %s",
-				guid);
-		goto err;
-	}
-
-	nelem = ploop_get_child_count_by_uuid(di, guid);
-	if (nelem == 0) {
-		if (strcmp(di->snapshots[snap_id]->parent_guid, NONE_UUID) == 0) {
-			ret = SYSEXIT_PARAM;
-			ploop_err(0, "Unable to delete base image");
-			goto err;
-		}
-		/* snapshot is not active and last -> delete */
-		ret = ploop_di_remove_image(di, guid, 1, &fname);
-		if (ret)
-			goto err;
-		get_disk_descriptor_fname(di, conf, sizeof(conf));
-		ret = ploop_store_diskdescriptor(conf, di);
-		if (ret)
-			goto err;
-		ploop_log(0, "Removing %s", fname);
-		if (fname != NULL && unlink(fname)) {
-			ploop_err(errno, "unlink %s", fname);
-			ret = SYSEXIT_UNLINK;
-		}
-		if (ret == 0)
-			ploop_log(0, "ploop snapshot %s has been successfully deleted",
-				guid);
-	} else if (nelem == 1) {
-		ret = ploop_merge_snapshot_by_guid(di, guid, PLOOP_MERGE_WITH_CHILD);
-	} else {
-		/* There no functionality to merge snapshot with >1 child */
-		ret = SYSEXIT_PARAM;
-		ploop_err(0, "There are %d references on %s snapshot: operation not supported",
-				nelem, guid);
-	}
-
-err:
-	free(fname);
-	ploop_unlock_di(di);
+	ploop_unlock_dd(di);
 
 	return ret;
 }
