@@ -39,10 +39,10 @@
 #define EXT4_SUPER_MAGIC	0xEF53
 #endif
 
-#define DISKDESCRIPTOR_XML	"DiskDescriptor.xml"
 /* Compatibility: Parallels use this UUID to mark top delta */
 #define TOPDELTA_UUID		"{5fbaabe3-6958-40ff-92a7-860e329aab41}"
 #define NONE_UUID		"{00000000-0000-0000-0000-000000000000}"
+#define UUID_SIZE		39	/* sizeof(TOPDELTA_UUID) for example */
 #define DEFAULT_FSTYPE		"ext4"
 #define BALLOON_FNAME		".balloon-c3a5ae3d-ce7f-43c4-a1ea-c61e2b4504e8"
 
@@ -58,11 +58,6 @@
 #define CHECK_DETAILED	0x08
 #define CHECK_REPAIR_SPARSE	0x10
 
-/* GPT */
-#define GPT_HEADER_SIZE		1
-#define GPT_PT_ENTRY_SIZE	32
-#define GPT_DATA_SIZE		(GPT_HEADER_SIZE + GPT_PT_ENTRY_SIZE)
-
 #define S2B(sec) ((off_t)(sec) << PLOOP1_SECTOR_LOG)
 #define B2S(sec) ((sec) >> PLOOP1_SECTOR_LOG)
 #define ROUNDUP(size, blocksize) \
@@ -228,6 +223,7 @@ static inline off_t ploop_ioff_to_sec(__u32 iblk, __u32 blocksize, int version)
 int gen_uuid_pair(char *uuid1, int len1, char *uuid2, int len2);
 int find_delta_names(const char * device, int start_level, int end_level,
 			    char **names, char ** format);
+PL_EXT int find_level_by_delta(const char *device, const char *delta);
 PL_EXT int ploop_get_attr(const char * device, const char * attr, int * res);
 int ploop_get_delta_attr(const char * device, int level, const char * attr, int * res);
 int ploop_get_size(const char * device, off_t * res);
@@ -283,9 +279,9 @@ int relocmap2relocblks(struct relocmap *relocmap, int lvl, __u32 a_h, __u32 n_sc
 			struct ploop_relocblks_ctl **relocblks_pp);
 PL_EXT int ploop_check(char *img, int flags, int ro, int raw, int verbose,
 		__u32 *blocksize_p);
-int check_deltas(struct ploop_disk_images_data *di, char **images,
-                struct ploop_mount_param *param, int raw, __u32 *blocksize);
-int ploop_check_delta(const char *image, int fd, __u64 blocksize);
+PL_EXT int check_deltas(struct ploop_disk_images_data *di, char **images,
+		int raw, __u32 *blocksize);
+PL_EXT int check_deltas_same(const char *img1, const char *img2);
 /* Logging */
 #define LOG_BUF_SIZE	8192
 int ploop_get_log_level(void);
@@ -314,25 +310,42 @@ void __ploop_err(int err_no, const char *format, ...)
 	 })
 
 char *make_sysfs_dev_name(int minor, char *buf, int len);
+int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *param, int flags);
 PL_EXT int ploop_mount(struct ploop_disk_images_data *di, char **images,
 		struct ploop_mount_param *param, int raw);
+PL_EXT int replace_delta(const char *device, int level, const char *image);
 PL_EXT int create_snapshot(const char *device, const char *delta, int syncfs);
 int get_list_size(char **list);
-void free_images_list(char **images);
+PL_EXT void free_images_list(char **images);
 int PWRITE(struct delta * delta, void * buf, unsigned int size, off_t off);
 int PREAD(struct delta * delta, void *buf, unsigned int size, off_t off);
 PL_EXT int ploop_getdevice(int *minor);
 struct ploop_disk_images_data *alloc_diskdescriptor(void);
+int read_diskdescriptor(const char *fname, struct ploop_disk_images_data *di);
 void get_disk_descriptor_fname(struct ploop_disk_images_data *di, char *buf, int size);
 void get_disk_descriptor_lock_fname(struct ploop_disk_images_data *di, char *out, int size);
 int find_image_idx_by_guid(struct ploop_disk_images_data *di, const char *guid);
-int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di, int check_state, char *out, int len);
+int ploop_find_dev(const char *component_name, const char *image, char *out, int size);
+int ploop_find_dev_by_cn(struct ploop_disk_images_data *di, const char *component_name,
+		int check_state, char *out, int len);
+int ploop_find_dev_by_dd(struct ploop_disk_images_data *di,
+		char *out, int len);
 int sys_fallocate(int fd, int mode, off_t offset, off_t len);
 int sys_syncfs(int fd);
+int create_snapshot_delta(const char *path, __u32 blocksize, off_t bdsize,
+		int version);
+int get_image_param_online(const char *device, off_t *size,
+		__u32 *blocksize, int *version);
+int get_image_param(struct ploop_disk_images_data *di, const char *guid,
+		off_t *size, __u32 *blocksize, int *version);
+int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
+                off_t *size, __u32 *blocksize, int *version);
+PL_EXT char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse);
 
 // manage struct ploop_disk_images_data
 int ploop_di_add_image(struct ploop_disk_images_data *di, const char *fname,
 		const char *guid, const char *parent_guid);
+void ploop_di_set_temporary(struct ploop_disk_images_data *di, const char *guid);
 int ploop_di_remove_image(struct ploop_disk_images_data *di, const char *guid,
 		int renew_top_uuid, char **fname);
 int ploop_di_merge_image(struct ploop_disk_images_data *di, const char *guid, char **fname);
@@ -341,8 +354,7 @@ PL_EXT char *find_image_by_guid(struct ploop_disk_images_data *di, const char *g
 PL_EXT int find_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_add_image_entry(struct ploop_disk_images_data *di, const char *fname, const char *guid);
 int ploop_add_snapshot_entry(struct ploop_disk_images_data *di, const char *guid,
-		const char *parent_guid);
-int ploop_find_dev(const char *module, const char *image, char *out, int size);
+		const char *parent_guid, int temporary);
 
 //balloon
 PL_EXT char *mntn2str(int mntn_type);
@@ -359,6 +371,10 @@ PL_EXT int ploop_discard_by_dev(const char *device, const char *mount_point,
 int ploop_blk_discard(const char* device, __u32 blocksize, off_t start, off_t end);
 
 /* lock */
+int ploop_lock_dd(struct ploop_disk_images_data *di);
+void ploop_unlock_dd(struct ploop_disk_images_data *di);
+int ploop_read_dd(struct ploop_disk_images_data *di);
+void ploop_clear_dd(struct ploop_disk_images_data *di);
 int ploop_lock_di(struct ploop_disk_images_data *di);
 void ploop_unlock_di(struct ploop_disk_images_data *di);
 int ploop_global_lock(void);
@@ -370,9 +386,10 @@ int make_fs(const char *device, const char *fstype, unsigned int fsblocksize);
 void tune_fs(const char *target, const char *device, unsigned long long size);
 int resize_fs(const char *device, off_t blocks);
 int dumpe2fs(const char *device, struct dump2fs_data *data);
-int e2fsck(const char *device, int flags);
+int e2fsck(const char *device, int flags, int *rc);
 int create_gpt_partition(const char *dev, off_t size, __u32 blocksize);
-int resize_gpt_partition(const char *devname, __u64 new_size);
+int resize_gpt_partition(const char *devname, __u64 new_size512, __u32 blocksize512);
+int check_and_repair_gpt(const char *device, __u32 blocksize512);
 
 // misc
 void get_basedir(const char *fname, char *out, int len);
@@ -397,16 +414,20 @@ int auto_mount_image(struct ploop_disk_images_data *di,
 void free_mount_param(struct ploop_mount_param *param);
 int check_and_restore_fmt_version(struct ploop_disk_images_data *di);
 int check_blockdev_size(unsigned long long sectors, __u32 blocksize, int version);
-
+int print_output(int level, const char *cmd, const char *arg);
+int read_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg);
+int write_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg);
+const char *get_snap_str(int temporary);
 // merge
 PL_EXT int get_delta_info(const char *device, struct merge_info *info);
 PL_EXT int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
 		char **images);
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode);
+int merge_temporary_snapshots(struct ploop_disk_images_data *di);
 
 PL_EXT int ploop_change_fmt_version(struct ploop_disk_images_data *di,
 		int new_version, int flags);
-int ploop_get_dev_by_delta(const char *delta, const char *component_name,
-		char **out[]);
+int ploop_get_dev_by_delta(const char *delta, const char *topdelta,
+		const char *component_name, char **out[]);
 
 #endif
