@@ -236,6 +236,8 @@ static int parse_xml(const char *basedir, xmlNode *root_node, struct ploop_disk_
 	cur_node = seek(root_node, "/Snapshots/Shot");
 	if (cur_node != NULL) {
 		for (; cur_node; cur_node = cur_node->next) {
+			int temporary = 0;
+
 			if (cur_node->type != XML_ELEMENT_NODE)
 				continue;
 
@@ -252,7 +254,11 @@ static int parse_xml(const char *basedir, xmlNode *root_node, struct ploop_disk_
 
 			ERR(parent_guid, "ParentGUID");
 
-			if (ploop_add_snapshot_entry(di, guid, parent_guid))
+			node = seek(cur_node, "Temporary");
+			if (node != NULL)
+				temporary = 1;
+
+			if (ploop_add_snapshot_entry(di, guid, parent_guid, temporary))
 				return -1;
 		}
 	}
@@ -314,7 +320,7 @@ static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 	if (di->nimages != di->nsnapshots) {
 		int ret;
 
-		ret = ploop_add_snapshot_entry(di, TOPDELTA_UUID, NONE_UUID);
+		ret = ploop_add_snapshot_entry(di, TOPDELTA_UUID, NONE_UUID, 0);
 		if (ret)
 			return ret;
 		if (di->top_guid == NULL)
@@ -326,11 +332,27 @@ static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 		 */
 		di->top_guid = strdup(TOPDELTA_UUID);
 	}
+
+
 	if (!is_valid_guid(di->top_guid)) {
 		ploop_err(0, "Validation of %s failed: invalid top delta %s",
 				di->runtime->xml_fname, di->top_guid);
 		return -1;
 	}
+
+	int i = find_snapshot_by_guid(di, di->top_guid);
+	if (i == -1) {
+		ploop_err(0, "Validation of %s failed: top delta %s is not found",
+				di->runtime->xml_fname, di->top_guid);
+		return -1;
+	}
+
+	if (di->snapshots[i]->temporary) {
+		ploop_err(0, "Validation of %s failed: top delta %s is temporary",
+				di->runtime->xml_fname, di->top_guid);
+		return -1;
+	}
+
 	if (di->nimages != di->nsnapshots) {
 		ploop_err(0, "Validation of %s failed: images(%d) != snapshots(%d)",
 				di->runtime->xml_fname, di->nimages, di->nsnapshots);
@@ -339,20 +361,24 @@ static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 	return 0;
 }
 
-static int read_diskdescriptor(const char *fname, struct ploop_disk_images_data *di)
+int ploop_read_dd(struct ploop_disk_images_data *di)
 {
 	int ret;
-	char path[PATH_MAX];
 	char basedir[PATH_MAX];
+	const char *fname;
 	struct stat st;
 	xmlDoc *doc = NULL;
 	xmlNode *root_element = NULL;
 
 	LIBXML_TEST_VERSION
 
-	if (di == NULL)
+	if (!di || !di->runtime || !di->runtime->xml_fname) {
+		ploop_err(0, "DiskDescriptor.xml is not opened");
 		return -1;
+	}
+	ploop_clear_dd(di);
 
+	fname = di->runtime->xml_fname;
 	if (stat(fname, &st)) {
 		ploop_err(errno, "Can't stat %s", fname);
 		return -1;
@@ -364,13 +390,6 @@ static int read_diskdescriptor(const char *fname, struct ploop_disk_images_data
 		return -1;
 	}
 
-	if (realpath(fname, path) == NULL) {
-		ploop_err(errno, "Can't resolve %s", fname);
-		return -1;
-	}
-
-	di->runtime->xml_fname = strdup(path);
-
 	doc = xmlReadFile(fname, NULL, 0);
 	if (doc == NULL) {
 		ploop_err(0, "Can't parse %s", fname);
@@ -378,7 +397,7 @@ static int read_diskdescriptor(const char *fname, struct ploop_disk_images_data
 	}
 	root_element = xmlDocGetRootElement(doc);
 
-	get_basedir(path, basedir, sizeof(path));
+	get_basedir(fname, basedir, sizeof(basedir));
 	ret = parse_xml(basedir, root_element, di);
 	if (ret == 0)
 		ret = validate_disk_descriptor(di);
@@ -388,19 +407,32 @@ static int read_diskdescriptor(const char *fname, struct ploop_disk_images_data
 	return ret;
 }
 
-int ploop_read_disk_descr(struct ploop_disk_images_data **di, const char *file)
+int read_diskdescriptor(const char *fname,
+		struct ploop_disk_images_data *di)
 {
-	*di = alloc_diskdescriptor();
-	if (*di == NULL)
-		return SYSEXIT_MALLOC;
+	char *path;
 
-	if (read_diskdescriptor(file, *di)) {
-		ploop_free_diskdescriptor(*di);
-		*di = NULL;
-		return SYSEXIT_DISKDESCR;
+	path = realpath(fname, NULL);
+	if (!path) {
+		ploop_err(errno, "Can't resolve %s", fname);
+		return -1;
 	}
 
-	return 0;
+	free(di->runtime->xml_fname);
+	di->runtime->xml_fname = path;
+
+	return ploop_read_dd(di);
+}
+
+int ploop_read_disk_descr(struct ploop_disk_images_data **di, const char *file)
+{
+	int ret;
+
+	ret = ploop_open_dd(di, file);
+	if (ret)
+		return ret;
+
+	return ploop_read_dd(*di);
 }
 
 static int normalize_image_name(const char *basedir, const char *image, char *out, int len)
@@ -431,6 +463,9 @@ int ploop_store_diskdescriptor(const char *fname, struct ploop_disk_images_data
 	if (convert_disk_descriptor(di))
 		return -1;
 
+	if (di->runtime->xml_fname == NULL)
+		di->runtime->xml_fname = strdup(fname);
+
 	get_basedir(fname, tmp, sizeof(tmp));
 	if (tmp[0] == '\0')
 		strcpy(tmp, "./");
@@ -638,6 +673,14 @@ int ploop_store_diskdescriptor(const char *fname, struct ploop_disk_images_data
 			goto err;
 		}
 
+		if (di->snapshots[i]->temporary) {
+			rc = xmlTextWriterWriteElement(writer, BAD_CAST "Temporary",  BAD_CAST "");
+			if (rc < 0) {
+				ploop_err(0, "Error at xmlTextWriter Temporary");
+				goto err;
+			}
+		}
+
 		/*  Close Shot */
 		rc = xmlTextWriterEndElement(writer);
 		if (rc < 0) {
