@@ -198,7 +198,6 @@ static int plooptool_mount(int argc, char **argv)
 {
 	int i, f, ret = 0;
 	int raw = 0;
-	int base = 0;
 	struct ploop_mount_param mountopts = {};
 	const char *component_name = NULL;
 
@@ -228,10 +227,10 @@ static int plooptool_mount(int argc, char **argv)
 			mountopts.fstype = strdup(optarg);
 			break;
 		case 'u':
-			if (!strcmp(optarg, "base"))
-				base = 1;
-			else
-				mountopts.guid = strdup(optarg);
+			mountopts.guid = parse_uuid(optarg);
+			if (!mountopts.guid)
+				return SYSEXIT_PARAM;
+
 			break;
 		case 'o':
 			mountopts.mount_data = strdup(optarg);
@@ -266,23 +265,16 @@ static int plooptool_mount(int argc, char **argv)
 	if (argc == 1 && is_xml_fname(argv[0]))
 	{
 		struct ploop_disk_images_data *di;
-		ret = read_dd(&di, argv[0]);
+		ret = ploop_open_dd(&di, argv[0]);
 		if (ret)
 			return ret;
 
 		if (component_name != NULL)
 			ploop_set_component_name(di, component_name);
-		if (base) {
-			mountopts.guid = ploop_get_base_delta_uuid(di);
-			if (mountopts.guid == NULL) {
-				ret = SYSEXIT_DISKDESCR;
-				fprintf(stderr, "Unable to find base delta uuid\n");
-				goto err;
-			}
-		}
+
 		ret = ploop_mount_image(di, &mountopts);
-err:
-		ploop_free_diskdescriptor(di);
+
+		ploop_close_dd(di);
 	}
 	else
 		ret = ploop_mount(NULL, argv, &mountopts, raw);
@@ -506,7 +498,7 @@ static int plooptool_umount(int argc, char **argv)
 		ret = ploop_umount(device, NULL);
 	} else if (is_xml_fname(argv[0])) {
 		struct ploop_disk_images_data *di;
-		ret = read_dd(&di, argv[0]);
+		ret = ploop_open_dd(&di, argv[0]);
 		if (ret)
 			return ret;
 
@@ -515,7 +507,7 @@ static int plooptool_umount(int argc, char **argv)
 
 		ret = ploop_umount_image(di);
 
-		ploop_free_diskdescriptor(di);
+		ploop_close_dd(di);
 	} else {
 		if (ploop_find_dev(component_name, argv[0], device, sizeof(device)) != 0) {
 			fprintf(stderr, "Image %s is not mounted\n", argv[0]);
@@ -610,7 +602,9 @@ static int plooptool_snapshot(int argc, char **argv)
 			syncfs = 1;
 			break;
 		case 'u':
-			param.guid = optarg;
+			param.guid = parse_uuid(optarg);
+			if (!param.guid)
+				return SYSEXIT_PARAM;
 			break;
 		default:
 			usage_snapshot();
@@ -628,13 +622,13 @@ static int plooptool_snapshot(int argc, char **argv)
 
 	if (is_xml_fname(argv[0])) {
 		struct ploop_disk_images_data *di;
-		ret = read_dd(&di, argv[0]);
+		ret = ploop_open_dd(&di, argv[0]);
 		if (ret)
 			return ret;
 
 		ret = ploop_create_snapshot(di, &param);
 
-		ploop_free_diskdescriptor(di);
+		ploop_close_dd(di);
 	} else {
 		if (!device) {
 			usage_snapshot();
@@ -646,6 +640,59 @@ static int plooptool_snapshot(int argc, char **argv)
 	return ret;
 }
 
+static void usage_tsnapshot(void)
+{
+	fprintf(stderr, "Usage: ploop tsnapshot -u <uuid> -c <component_name>\n"
+			"	[-m MOUNT_POINT] DiskDescriptor.xml\n"
+		);
+}
+
+static int plooptool_tsnapshot(int argc, char **argv)
+{
+	int i, ret;
+	struct ploop_disk_images_data *di;
+	struct ploop_tsnapshot_param param = {};
+
+	while ((i = getopt(argc, argv, "u:c:m:")) != EOF) {
+		switch (i) {
+		case 'u':
+			param.guid = parse_uuid(optarg);
+			if (!param.guid)
+				return SYSEXIT_PARAM;
+			break;
+		case 'c':
+			param.component_name = optarg;
+			break;
+		case 'm':
+			param.target = optarg;
+			break;
+		default:
+			usage_snapshot();
+			return SYSEXIT_PARAM;
+		}
+	}
+
+	argc -= optind;
+	argv += optind;
+
+	if (argc != 1 || !is_xml_fname(argv[0]) ||
+			param.guid == NULL ||
+			param.component_name == NULL) {
+		usage_tsnapshot();
+		return SYSEXIT_PARAM;
+	}
+
+	ret = ploop_open_dd(&di, argv[0]);
+	if (ret)
+		return ret;
+
+	ret = ploop_create_temporary_snapshot(di, &param, NULL);
+
+	ploop_close_dd(di);
+
+	return ret;
+}
+
 static void usage_snapshot_switch(void)
 {
 	fprintf(stderr, "Usage: ploop snapshot-switch -u <uuid> DiskDescriptor.xml\n"
@@ -662,7 +709,9 @@ static int plooptool_snapshot_switch(int argc, char **argv)
 	while ((i = getopt(argc, argv, "u:D")) != EOF) {
 		switch (i) {
 		case 'u':
-			uuid = optarg;
+			uuid = parse_uuid(optarg);
+			if (!uuid)
+				return SYSEXIT_PARAM;
 			break;
 		case 'D':
 			/* for test purposes */
@@ -682,13 +731,13 @@ static int plooptool_snapshot_switch(int argc, char **argv)
 		return SYSEXIT_PARAM;
 	}
 
-	ret = read_dd(&di, argv[0]);
+	ret = ploop_open_dd(&di, argv[0]);
 	if (ret)
 		return ret;
 
 	ret = ploop_switch_snapshot(di, uuid, flags);
 
-	ploop_free_diskdescriptor(di);
+	ploop_close_dd(di);
 
 	return ret;
 }
@@ -708,7 +757,9 @@ static int plooptool_snapshot_delete(int argc, char **argv)
 	while ((i = getopt(argc, argv, "u:")) != EOF) {
 		switch (i) {
 		case 'u':
-			uuid = optarg;
+			uuid = parse_uuid(optarg);
+			if (!uuid)
+				return SYSEXIT_PARAM;
 			break;
 		default:
 			usage_snapshot_delete();
@@ -724,13 +775,13 @@ static int plooptool_snapshot_delete(int argc, char **argv)
 		return SYSEXIT_PARAM;
 	}
 
-	ret = read_dd(&di, argv[0]);
+	ret = ploop_open_dd(&di, argv[0]);
 	if (ret)
 		return ret;
 
 	ret = ploop_delete_snapshot(di, uuid);
 
-	ploop_free_diskdescriptor(di);
+	ploop_close_dd(di);
 
 	return ret;
 }
@@ -749,7 +800,9 @@ static int plooptool_snapshot_merge(int argc, char ** argv)
 	while ((i = getopt(argc, argv, "u:A")) != EOF) {
 		switch (i) {
 		case 'u':
-			param.guid = optarg;
+			param.guid = parse_uuid(optarg);
+			if (!param.guid)
+				return SYSEXIT_PARAM;
 			break;
 		case 'A':
 			param.merge_all = 1;
@@ -771,13 +824,13 @@ static int plooptool_snapshot_merge(int argc, char ** argv)
 
 	if (argc == 1 && is_xml_fname(argv[0])) {
 		struct ploop_disk_images_data *di;
-		ret = read_dd(&di, argv[0]);
+		ret = ploop_open_dd(&di, argv[0]);
 		if (ret)
 			return ret;
 
 		ret = ploop_merge_snapshot(di, &param);
 
-		ploop_free_diskdescriptor(di);
+		ploop_close_dd(di);
 	} else {
 		usage_snapshot_merge();
 		return SYSEXIT_PARAM;
@@ -857,13 +910,13 @@ static int plooptool_resize(int argc, char **argv)
 		return SYSEXIT_PARAM;
 	}
 
-	ret = read_dd(&di, argv[0]);
+	ret = ploop_open_dd(&di, argv[0]);
 	if (ret)
 		return ret;
 
 	ret = ploop_resize_image(di, &param);
 
-	ploop_free_diskdescriptor(di);
+	ploop_close_dd(di);
 
 	return ret;
 }
@@ -912,7 +965,7 @@ static int plooptool_convert(int argc, char **argv)
 		return SYSEXIT_PARAM;
 	}
 
-	ret = read_dd(&di, argv[0]);
+	ret = ploop_open_dd(&di, argv[0]);
 	if (ret)
 		return ret;
 	if (mode != -1)
@@ -920,7 +973,7 @@ static int plooptool_convert(int argc, char **argv)
 	else if (version != -1)
 		ret = ploop_change_fmt_version(di, version, 0);
 
-	ploop_free_diskdescriptor(di);
+	ploop_close_dd(di);
 
 	return ret;
 }
@@ -973,7 +1026,7 @@ static int plooptool_info(int argc, char **argv)
 		struct ploop_disk_images_data *di;
 		struct ploop_spec spec = {};
 
-		ret = read_dd(&di, argv[0]);
+		ret = ploop_open_dd(&di, argv[0]);
 		if (ret)
 			return ret;
 
@@ -984,7 +1037,7 @@ static int plooptool_info(int argc, char **argv)
 				spec.blocksize,
 				spec.fmt_version);
 
-		ploop_free_diskdescriptor(di);
+		ploop_close_dd(di);
 	} else {
 		ret = ploop_get_info_by_descr(argv[0], &info);
 		if (ret == 0)
@@ -1065,9 +1118,170 @@ static int plooptool_list(int argc, char **argv)
 	return 0;
 }
 
+static void usage_replace(void)
+{
+	fprintf(stderr, "Usage: ploop replace {-d DEVICE | -m MNT} {-l LVL | -o СDELTA} -i DELTA\n"
+			"       ploop replace {-u UUID | -l LVL | -o CDELTA} -i DELTA DiskDescriptor.xml\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
+			"       MNT := directory where ploop is mounted to\n"
+			"       LVL := NUMBER, distance from base delta\n"
+			"       UUID := UUID of image to be replaced\n"
+			"       CDELTA := path to currently used image file\n"
+			"       DELTA := path to new image file\n"
+	       );
+}
+
+
+static int plooptool_replace(int argc, char **argv)
+{
+	int i;
+	char dev[PATH_MAX];
+	char *device = NULL;
+	char *mnt = NULL;
+	struct ploop_replace_param param = {
+		.level = -1,
+	};
+
+	while ((i = getopt(argc, argv, "d:m:l:i:u:o:")) != EOF) {
+		switch (i) {
+		case 'd':
+			device = optarg;
+			break;
+		case 'm':
+			mnt = optarg;
+			break;
+		case 'l':
+			param.level = atoi(optarg);
+			break;
+		case 'u':
+			param.guid = parse_uuid(optarg);
+			if (!param.guid)
+				return SYSEXIT_PARAM;
+			break;
+		case 'i':
+			param.file = strdup(optarg);
+			break;
+		case 'o':
+			param.cur_file = strdup(optarg);
+			break;
+		default:
+			usage_replace();
+			return SYSEXIT_PARAM;
+		}
+	}
+
+	argc -= optind;
+	argv += optind;
+
+	if (!param.file) {
+		fprintf(stderr, "Error: image file not specified (use -i)\n");
+		usage_replace();
+		return SYSEXIT_PARAM;
+	}
+
+	if ((argc == 1) && is_xml_fname(argv[0])) {
+		int ret;
+		struct ploop_disk_images_data *di;
+
+		/* only one way of choosing delta to replace */
+		if ( (!!param.guid) + (param.level != -1) +
+				(!!param.cur_file) != 1)  {
+			fprintf(stderr, "Error: either one of uuid (-u), "
+					"level (-l) or current file (-o) "
+					"must be specified\n");
+			usage_replace();
+			return SYSEXIT_PARAM;
+		}
+
+		ret = ploop_open_dd(&di, argv[0]);
+		if (ret)
+			return ret;
+
+		ret = ploop_replace_image(di, &param);
+		ploop_close_dd(di);
+
+		return ret;
+	}
+	else {
+		int level = param.level;
+
+		if (argc > 0) {
+			usage_replace();
+			return SYSEXIT_PARAM;
+		}
+		if ((!!device) + (!!mnt) != 1) {
+			fprintf(stderr, "Error: either device (-d), mount "
+					"point (-m) or DiskDescriptor.xml "
+					"must be specified\n");
+			usage_replace();
+			return SYSEXIT_PARAM;
+		}
+		if (mnt) {
+			if (ploop_get_dev_by_mnt(mnt, dev, sizeof(dev))) {
+				fprintf(stderr, "Unable to find ploop device "
+						"by mount point %s\n", mnt);
+				return SYSEXIT_PARAM;
+			}
+			device = dev;
+		}
+		/* Either level or current delta must be specified */
+		if ((level != -1) + (!!param.cur_file) != 1) {
+			fprintf(stderr, "Error: either one of level (-l) or "
+					"current delta file (-o) must be "
+					"specified\n");
+			usage_replace();
+			return SYSEXIT_PARAM;
+		}
+		if (param.cur_file) {
+			int ret;
+
+			level = find_level_by_delta(device, param.cur_file);
+			if (level < 0) {
+				fprintf(stderr, "Can't find level by "
+						"delta file name %s",
+						param.cur_file);
+				return SYSEXIT_PARAM;
+			}
+
+			ret = check_deltas_same(param.cur_file, param.file);
+			if (ret)
+				return ret;
+		}
+
+		return replace_delta(device, level, param.file);
+	}
+}
+
 int main(int argc, char **argv)
 {
 	char * cmd;
+	int v = 3;
+
+	/* global options */
+	while (argc > 1 && argv[1][0] == '-') {
+		switch (argv[1][1]) {
+			case 'v':
+				switch (argv[1][2]) {
+				    case '\0':
+					v++;
+					break;
+				    case 'v': /* -vvv... */
+					v += strlen(&argv[1][1]);
+					break;
+				    default: /* -vNN */
+					v = atoi(&argv[1][2]);
+				}
+				break;
+			case '-': /* long option */
+				/* fall through */
+			default:
+				fprintf(stderr, "Bad option %s\n", argv[1]);
+				usage_summary();
+				return SYSEXIT_PARAM;
+		}
+		argc--;
+		argv++;
+	}
 
 	if (argc < 2) {
 		usage_summary();
@@ -1078,7 +1292,7 @@ int main(int argc, char **argv)
 	argc--;
 	argv++;
 
-	ploop_set_verbose_level(3);
+	ploop_set_verbose_level(v);
 	init_signals();
 
 	if (strcmp(cmd, "init") == 0)
@@ -1097,6 +1311,8 @@ int main(int argc, char **argv)
 		return plooptool_rm(argc, argv);
 	if (strcmp(cmd, "snapshot") == 0)
 		return plooptool_snapshot(argc, argv);
+	if (strcmp(cmd, "tsnapshot") == 0)
+		return plooptool_tsnapshot(argc, argv);
 	if (strcmp(cmd, "snapshot-switch") == 0)
 		return plooptool_snapshot_switch(argc, argv);
 	if (strcmp(cmd, "snapshot-delete") == 0)
@@ -1130,6 +1346,8 @@ int main(int argc, char **argv)
 		return plooptool_stat(argc, argv);
 	if (strcmp(cmd, "copy") == 0)
 		return plooptool_copy(argc, argv);
+	if (strcmp(cmd, "replace") == 0)
+		return plooptool_replace(argc, argv);
 
 	if (cmd[0] != '-') {
 		char ** nargs;
